import locale
import logging
from typing import Optional

from numpy import ndarray
from pandas import Series, to_numeric
from pandas.api.types import is_numeric_dtype

from ..schema import InferredField
from .base import BaseType

logger = logging.getLogger(__name__)


class FloatType(BaseType):
    weight = 8

    def validate(self, data: ndarray) -> Optional[InferredField]:
        for value in data:
            try:
                if isinstance(value, str):
                    float(locale.atof(value))
                else:
                    float(value)
            except Exception:  # noqa
                logger.debug(f"Cannot cast the value '{value}' to {self.name}")
                return None
        return InferredField(inferred_type=self.name)

    def fix(self, column: Series, inferred_field: InferredField) -> Series:
        if is_numeric_dtype(column):
            return column
        return to_numeric(column.apply(locale.atof))


def get_type() -> FloatType:
    return FloatType()



from typing import Optional

from numpy import ndarray
from pandas import Series, to_datetime

from ..schema import InferredField
from . import validate_strptime_pattern
from .base import BaseType


class DateType(BaseType):
    weight = 2
    patterns: list[str] = [
        "%m/%d/%Y",  # 12/01/2022
        "%m-%d-%Y",  # 12-01-2022
        "%m.%d.%Y",  # 12.01.2022
        "%m/%d/%y",  # 12/01/22
        "%m-%d-%y",  # 12.01.2022
        "%m.%d.%y",  # 12.01.22
        "%Y/%m/%d",  # 2022/12/01
        "%Y-%m-%d",  # 2022-12-01
        "%Y.%m.%d",  # 2022.12.01
        "%d/%m/%Y",  # 01/12/2022
        "%d-%m-%Y",  # 01-12-2022
        "%d.%m.%Y",  # 01.12.2022
        "%d/%b/%Y",  # 01/Mar/2022
        "%d-%b-%Y",  # 01-Mar-2022
        "%Y-%m",  # 2022-07
    ]

    def validate(self, data: ndarray) -> Optional[InferredField]:
        for pattern in self.patterns:
            if validate_strptime_pattern(data, pattern):
                return InferredField(inferred_type=self.name, inferred_pattern=pattern)

    def fix(self, column: Series, inferred_field: InferredField) -> Series:
        return to_datetime(column, format=inferred_field.inferred_pattern)


def get_type() -> DateType:
    return DateType()



import logging
from datetime import datetime

from numpy import ndarray

logger = logging.getLogger(__name__)


def validate_strptime_pattern(data: ndarray, pattern: str) -> bool:
    validated: bool = False
    for value in data:
        try:
            if isinstance(value, str):  # For security reasons & skip nan values
                datetime.strptime(value, pattern)
                validated = True
        except ValueError:
            logger.debug(f"Cannot cast the value '{value}' using pattern '{pattern}'")
            return False
    return validated



from typing import Optional, Tuple

import numpy as np
from numpy import ndarray
from pandas import Series

from ..schema import InferredField
from .base import BaseType


class BooleanType(BaseType):
    weight = 1
    patterns: list[Tuple] = [  # The first member should be the true value
        ("YES", "NO"),
        ("TRUE", "FALSE"),
        ("T", "F"),
        ("Y", "N"),
        (1, 0),
    ]

    @staticmethod
    def _validate_value_pattern(data: ndarray, pattern: Tuple) -> bool:
        match_first_value: bool = False
        match_second_value: bool = False

        for value in data:
            if isinstance(value, str):
                value = value.lower()
            if isinstance(pattern[0], str):
                pattern = (pattern[0].lower(), pattern[1].lower())
            if value == pattern[0]:
                match_first_value = True
            elif value == pattern[1]:
                match_second_value = True
            else:
                return False
        if match_first_value and match_second_value:
            return True
        return False

    def validate(self, data: ndarray) -> Optional[InferredField]:
        if data.dtype == np.dtype("bool"):
            return InferredField(inferred_type=self.name)
        for pattern in self.patterns:
            if self._validate_value_pattern(data, pattern):
                return InferredField(inferred_type=self.name, inferred_pattern=pattern)

    def fix(self, column: Series, inferred_field: InferredField) -> Series:
        if not inferred_field.inferred_pattern:
            raise NotImplementedError()  # Missing pattern - Mean that the column recognized as boolean by pandas

        if isinstance(inferred_field.inferred_pattern[0], str):
            return column.apply(
                lambda x: True
                if x.lower() == inferred_field.inferred_pattern[0].lower()
                else False
            )

        return column.apply(
            lambda x: True if x == inferred_field.inferred_pattern[0] else False
        )


def get_type() -> BooleanType:
    return BooleanType()



import logging
import os
from importlib import import_module
from types import ModuleType
from typing import Any, Dict, List, Optional, Type

from numpy import ndarray
from numpy.dtypes import (
    BoolDType,
    DateTime64DType,
    Float16DType,
    Float32DType,
    Float64DType,
    Int8DType,
    Int16DType,
    Int32DType,
    Int64DType,
    IntDType,
    ObjectDType,
    StrDType,
)
from pandas import Series

from .schema import Field
from .types.base import BaseType
from .types.boolean import BooleanType
from .types.datetime_ import DatetimeType
from .types.float import FloatType
from .types.integer import IntegerType
from .types.string import StringType

logger = logging.getLogger(__name__)


class TypesHandler:
    PACKAGE_NAME: str = "shmessy"
    TYPES_DIR: str = "types"

    def __init__(self):
        self.__types = self._discover_types()
        self.__types_as_dict: Dict[str, BaseType] = self._types_as_dict(self.__types)

    @classmethod
    def _types_as_dict(cls, __types: List[BaseType]) -> Dict[str, BaseType]:
        res = {}
        for type_ in __types:
            res[type_.name] = type_
        return res

    @classmethod
    def _discover_types(cls) -> List[BaseType]:
        types: List[BaseType] = []
        root_directory = os.path.join(os.path.dirname(__file__))
        types_directory = os.path.join(root_directory, cls.TYPES_DIR)

        for filename in os.listdir(types_directory):
            try:
                types.append(cls._load_type(filename).get_type())
            except AttributeError:
                pass  # ignore types without factory (Such as base types)
        return types

    @classmethod
    def _load_type(cls, type_filename: str) -> Optional[ModuleType]:
        module = (
            f"{cls.PACKAGE_NAME}.{cls.TYPES_DIR}.{type_filename.replace('.py', '')}"
        )
        try:
            return import_module(module)
        except (ImportError, ValueError, AttributeError) as e:
            logger.error(f"Couldn't import {module}: {e}")

    def fix_field(self, column: Series, inferred_field: Field) -> Any:
        try:
            fixed_field = self.__types_as_dict[inferred_field.inferred_type].fix(
                column=column, inferred_field=inferred_field
            )
            if fixed_field is not None:
                return fixed_field
        except NotImplementedError:
            pass

        return column

    def infer_field(self, field_name: str, data: ndarray) -> Field:
        sorted_types = sorted(self.__types, key=lambda x: x.weight)
        for type_ in sorted_types:
            logger.debug(f"Trying to match column {field_name} to type {type_.name}")
            inferred = type_.validate(data)
            if inferred:
                return Field(
                    field_name=field_name,
                    source_type=_numpy_type_shmessy_type(data.dtype),
                    inferred_type=inferred.inferred_type,
                    inferred_pattern=inferred.inferred_pattern,
                )

        return Field(
            field_name=field_name, source_type=_numpy_type_shmessy_type(data.dtype)
        )


def _numpy_type_shmessy_type(numpy_type: Type) -> str:
    if isinstance(
        numpy_type,
        (
            IntDType,
            Int64DType,
            Int8DType,
            Int16DType,
            Int32DType,
        ),
    ):
        return IntegerType().name
    if isinstance(numpy_type, (ObjectDType, StrDType)):
        return StringType().name
    if isinstance(numpy_type, BoolDType):
        return BooleanType().name
    if isinstance(numpy_type, DateTime64DType):
        return DatetimeType().name
    if isinstance(numpy_type, (Float16DType, Float32DType, Float64DType)):
        return FloatType().name
    return str(numpy_type)
